#pragma once

#include "stdafx.h"
#include <sstream>  
#include <math.h>
#include <windows.h>
#include <string>
#include <algorithm>

#ifndef __COMPORT_H_
#define __COMPORT_H_

typedef unsigned char comportBits;

#undef PARITY_NONE
#undef PARITY_ODD
#undef PARITY_EVEN
#undef PARITY_MARK
#undef PARITY_SPACE
typedef enum { // as xxxPARITY in winbase.h
  PARITY_NONE  = 0x00,
  PARITY_ODD   = 0x01,
  PARITY_EVEN  = 0x02,
  PARITY_MARK  = 0x03,
  PARITY_SPACE = 0x04,
} comportParity;

typedef enum { // as xxxSTOPBIT(S) in winbase.h
  STOP_1   = 0x00,
  STOP_1_5 = 0x01,
  STOP_2   = 0x02
} comportStop;

typedef enum { // as DTR_CONTROL_xxx in winbase.h
  DTR_LOW       = 0x00,
  DTR_HIGH      = 0x01,
  DTR_HANDSHAKE = 0x02
} comportDTR;

typedef enum { // as RTS_CONTROL_xxx in winbase.h
  RTS_LOW       = 0x00,
  RTS_HIGH      = 0x01,
  RTS_HANDSHAKE = 0x02,
  RTS_TOGGLE    = 0x03
} comportRTS;

class ComPort {  // com port class code stolen from a thread on Hydrogen audio http://www.hydrogenaud.io/forums/index.php?showtopic=59380
  private:
    HANDLE hComHandle;

  protected:
    void init(const char *port, int baudrate, comportBits bits, comportParity parity,
              comportStop stop, comportDTR dtr, comportRTS rts);
  public:
    ComPort(const char *port);
    ComPort(const char *port, int baudrate);
    ComPort(const char *port, int baudrate, comportBits bits, comportParity parity,
            comportStop stop, comportDTR dtr, comportRTS rts);
    ~ComPort();

    void setBaud(int baudrate);
    void setParity(comportParity parity);
    void setBits(comportBits bits);
    void setStop(comportStop stop);
    void setDTR(comportDTR dtr);
    void setRTS(comportRTS cts);
    void txChar(unsigned char byte);
    void txStringN(int len, const char *str);
    void txString(const char *str);
    int rxStatus();
    unsigned char rxChar();
};


ComPort::ComPort(const char *port)
{
  init(port,9600,8,PARITY_NONE,STOP_2,DTR_HIGH,RTS_HIGH);
}

ComPort::ComPort(const char *port, int baudrate)
{
  init(port,baudrate,8,PARITY_NONE,STOP_2,DTR_HIGH,RTS_HIGH);
}

ComPort::ComPort(const char *port, int baudrate, comportBits bits, comportParity parity,
                 comportStop stop, comportDTR dtr, comportRTS rts)
{
  init(port,baudrate,bits,parity,stop,dtr,rts);
}

void ComPort::init(const char *port, int baudrate, comportBits bits, comportParity parity,
                   comportStop stop, comportDTR dtr, comportRTS rts)
{
  if (!port) {
    hComHandle = INVALID_HANDLE_VALUE;
    return;
  }

  char device[MAX_PATH];
  strcpy(device,"\\\\.\\"); strcat(device,port);
  hComHandle = CreateFileA(device,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
  if (hComHandle == INVALID_HANDLE_VALUE)
    throw exception_win32(GetLastError());

  if (!SetupComm(hComHandle,4096,1024)) {
    DWORD err = GetLastError();
    CloseHandle(hComHandle); hComHandle = INVALID_HANDLE_VALUE;
    throw exception_win32(err);
  }

  DCB dcb;
  memset(&dcb,0,sizeof(DCB));
  dcb.DCBlength = sizeof(DCB);
  dcb.BaudRate = baudrate;
  dcb.fBinary = TRUE;
  dcb.fParity = (parity == PARITY_NONE) ? 0 : 1;
  dcb.ByteSize = bits;
  dcb.Parity = parity;
  dcb.StopBits = stop;
  dcb.fDtrControl = dtr;
  dcb.fRtsControl = rts;
  if (!SetCommState(hComHandle,&dcb)) {
    DWORD err = GetLastError();
    CloseHandle(hComHandle); hComHandle = INVALID_HANDLE_VALUE;
    throw exception_win32(err);
  }

  COMMTIMEOUTS ctmo;
  ctmo.ReadIntervalTimeout = 0;
  ctmo.ReadTotalTimeoutMultiplier = 0;
  ctmo.ReadTotalTimeoutConstant = 0;
  ctmo.WriteTotalTimeoutMultiplier = 0;
  ctmo.WriteTotalTimeoutConstant = 0;
  if (!SetCommTimeouts(hComHandle,&ctmo)) {
    DWORD err = GetLastError();
    CloseHandle(hComHandle); hComHandle = INVALID_HANDLE_VALUE;
    throw exception_win32(err);
  }
}

ComPort::~ComPort()
{
  if (hComHandle != INVALID_HANDLE_VALUE)
    CloseHandle(hComHandle);
}

void ComPort::setBaud(int baudrate)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DCB dcb;
  dcb.DCBlength = sizeof(DCB);
  GetCommState(hComHandle,&dcb);
  dcb.BaudRate = baudrate;
  SetCommState(hComHandle,&dcb);
}

void ComPort::setBits(comportBits bits)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DCB dcb;
  dcb.DCBlength = sizeof(DCB);
  GetCommState(hComHandle,&dcb);
  dcb.ByteSize = bits;
  SetCommState(hComHandle,&dcb);
}

void ComPort::setParity(comportParity parity)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DCB dcb;
  dcb.DCBlength = sizeof(DCB);
  GetCommState(hComHandle,&dcb);
  dcb.Parity = parity;
  SetCommState(hComHandle,&dcb);
}

void ComPort::setStop(comportStop stop)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DCB dcb;
  dcb.DCBlength = sizeof(DCB);
  GetCommState(hComHandle,&dcb);
  dcb.StopBits = stop;
  SetCommState(hComHandle,&dcb);
}

void ComPort::setDTR(comportDTR dtr)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DCB dcb;
  dcb.DCBlength = sizeof(DCB);
  GetCommState(hComHandle,&dcb);
  dcb.fDtrControl = dtr;
  SetCommState(hComHandle,&dcb);
}

void ComPort::setRTS(comportRTS rts)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DCB dcb;
  dcb.DCBlength = sizeof(DCB);
  GetCommState(hComHandle,&dcb);
  dcb.fRtsControl = rts;
  SetCommState(hComHandle,&dcb);
}

void ComPort::txChar(unsigned char byte)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DWORD nb;
  WriteFile(hComHandle,&byte,1,&nb,NULL);
}

void ComPort::txStringN(int len, const char *str)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DWORD nb;
  WriteFile(hComHandle,str,len,&nb,NULL);
}

void ComPort::txString(const char *str)
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return;

  DWORD nb;
  WriteFile(hComHandle,str,strlen(str),&nb,NULL);
}

int ComPort::rxStatus()
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return 0;

  // TODO: implement - rx not required for this purpose
  return 0;
}

unsigned char ComPort::rxChar()
{
  if (hComHandle == INVALID_HANDLE_VALUE)
    return 0;

  // TODO: implement - rx not required for this purpose
  return 0;
}

#endif

std::string StringToUpper(std::string strToConvert)
{
    std::transform(strToConvert.begin(), strToConvert.end(), strToConvert.begin(), ::toupper);

    return strToConvert;
}

bool replaceAll(std::string& str, const std::string& from, const std::string& to) {
	if(from.empty())
        return false;
    size_t start_pos = 0;
    while((start_pos = str.find(from, start_pos)) != std::string::npos) {
        str.replace(start_pos, from.length(), to);
        start_pos += to.length(); // In case 'to' contains 'from', like replacing 'x' with 'yx'
    }
}

class foo_com_port_param: public cfg_string
{
public:
	explicit inline foo_com_port_param(const GUID & p_guid,const char * p_defaultval) : cfg_string(p_guid,pfc::string8(p_defaultval))   {}
};

// {E297AC59-B454-468F-9145-AA217AA9C0B5}
static const GUID foo_comm_branch_GUID = 
{ 0xe297ac59, 0xb454, 0x468f, { 0x91, 0x45, 0xaa, 0x21, 0x7a, 0xa9, 0xc0, 0xb5 } };
static advconfig_branch_factory foo_comm_branch("foo_comm", foo_comm_branch_GUID, advconfig_branch::guid_branch_tools, 0);

// {4AC58018-09D2-48DB-BCAD-AAF476471F7F}
static const GUID foo_comm_comport_GUID =
{ 0x4ac58018, 0x9d2, 0x48db, { 0xbc, 0xad, 0xaa, 0xf4, 0x76, 0x47, 0x1f, 0x7f } };
static advconfig_string_factory foo_comm_comport("Com port", foo_comm_comport_GUID, foo_comm_branch_GUID, 1, "COM1");

// {B03DDE71-B377-43AA-8C75-7AA2F7071581}
static const GUID foo_comm_combaud_GUID = 
{ 0xb03dde71, 0xb377, 0x43aa, { 0x8c, 0x75, 0x7a, 0xa2, 0xf7, 0x7, 0x15, 0x81 } };
static advconfig_integer_factory foo_comm_combaud("Baud rate", foo_comm_combaud_GUID, foo_comm_branch_GUID, 2, 9600,0,2000000);

// {C494D8CB-A5E3-43E3-9E51-D3FB90E7AF71}
static const GUID foo_comm_combits_GUID = 
{ 0xc494d8cb, 0xa5e3, 0x43e3, { 0x9e, 0x51, 0xd3, 0xfb, 0x90, 0xe7, 0xaf, 0x71 } };
static advconfig_integer_factory foo_comm_combits("Data bits", foo_comm_combits_GUID, foo_comm_branch_GUID, 3, 8,4,8);
// {751B2EFB-800C-4058-B7C1-3E910513CF19}
static const GUID foo_comm_comstop_GUID = 
{ 0x751b2efb, 0x800c, 0x4058, { 0xb7, 0xc1, 0x3e, 0x91, 0x5, 0x13, 0xcf, 0x19 } };
static advconfig_string_factory foo_comm_comstop("Stop bit", foo_comm_comstop_GUID, foo_comm_branch_GUID, 4, "1");

// {7C9FEDEC-A314-4ECA-85D4-41CCBB2DFCDE}
static const GUID foo_comm_comparity_GUID = 
{ 0x7c9fedec, 0xa314, 0x4eca, { 0x85, 0xd4, 0x41, 0xcc, 0xbb, 0x2d, 0xfc, 0xde } };
static advconfig_string_factory foo_comm_comparity("Parity", foo_comm_comparity_GUID, foo_comm_branch_GUID, 5, "n");

// {5B5546B8-1C4C-4A8A-9F53-FFEC2AD728F4}
static const GUID foo_comm_refreshtime_GUID = 
{ 0x5b5546b8, 0x1c4c, 0x4a8a, { 0x9f, 0x53, 0xff, 0xec, 0x2a, 0xd7, 0x28, 0xf4 } };
static advconfig_integer_factory foo_comm_refreshtime("Refresh interval (sec)", foo_comm_refreshtime_GUID, foo_comm_branch_GUID, 6, 1,1,2000000);

// {81A3955F-7F78-45A8-AA23-BD028FD1336B}
static const GUID foo_comm_lcdrefresh_GUID = 
{ 0x81a3955f, 0x7f78, 0x45a8, { 0xaa, 0x23, 0xbd, 0x2, 0x8f, 0xd1, 0x33, 0x6b } };
static advconfig_checkbox_factory foo_comm_lcdrefresh("LCD refresh", foo_comm_lcdrefresh_GUID, foo_comm_branch_GUID, 6, false);

// {A10073D0-70EC-415B-9F02-16C9A7982AAC}
static const GUID foo_comm_outputformat_GUID = 
{ 0xa10073d0, 0x70ec, 0x415b, { 0x9f, 0x2, 0x16, 0xc9, 0xa7, 0x98, 0x2a, 0xac } };
static advconfig_string_factory foo_comm_outputformat("Output format", foo_comm_outputformat_GUID, foo_comm_branch_GUID, 7, 
	  "$upper($left($pad(%filename%,$ifequal(%volume%,0,67,59)),$ifequal(%volume%,0,67,59))) $upper($left($pad(%codec%,3),3))/$left($pad(%bitrate%,3),3)/$ifequal(%volume%,0,,$left(%volume%,5)dB/)%playback_time%");

// {789C1D75-2072-418A-B713-9004B0BB5EBD}
static const GUID foo_comm_consoledebug_GUID = 
{ 0x789c1d75, 0x2072, 0x418a, { 0xb7, 0x13, 0x90, 0x4, 0xb0, 0xbb, 0x5e, 0xbd } };
static advconfig_checkbox_factory foo_comm_consoledebug("Console debug output", foo_comm_consoledebug_GUID, foo_comm_branch_GUID, 8, false);


class foo_comm : public play_callback_static 
{
public:
	// play_callback methods
	virtual void FB2KAPI on_playback_starting(play_control::t_track_command p_command,bool p_paused)
	{
	}
	virtual void FB2KAPI on_playback_new_track(metadb_handle_ptr p_track)
	{
		commecho(true);	
	}
	virtual void FB2KAPI on_playback_stop(play_control::t_stop_reason p_reason)
	{			
	}
	virtual void FB2KAPI on_playback_seek(double p_time)
	{
		commecho(true);			// don't send EOT to clear lcd during playback because it flickers				
	}
	virtual void FB2KAPI on_playback_pause(bool p_state)
	{
		commecho(true);						
	}
	virtual void FB2KAPI on_playback_edited(metadb_handle_ptr p_track){}
	virtual void FB2KAPI on_playback_dynamic_info(const file_info & p_info){}
	virtual void FB2KAPI on_playback_dynamic_info_track(const file_info &p_info){}
	virtual void FB2KAPI on_playback_time(double p_time)
	{
		int time=floor(p_time);
		if ((time % foo_comm_refreshtime)==0) commecho(false);  
	}
	virtual void FB2KAPI on_volume_change(float p_new_val)
	{
		commecho(true );							
	}
public:
	virtual unsigned get_flags()
	{
		return 65535; // handle all events
	}
	
	// echoes formated current track title to com port
	virtual void commecho(bool resetlcd)
	{
		pfc::string8 formatspec_str8;
		foo_comm_outputformat.get(formatspec_str8);
		std::string formatspec;
		formatspec=formatspec_str8.toString();
		std::stringstream out;
		metadb_handle_ptr p_track;
		static_api_ptr_t<playback_control> pc;
		if (pc->get_now_playing(p_track)) {

			long double vol=pc->get_volume();
			replaceAll(formatspec,"%volume%",std::to_string(vol));	// there is no %volume% in foobar title formating, so we have to do it manually

			pfc::string8 title;										
			service_ptr_t<titleformat_object> format;
			if (static_api_ptr_t<titleformat_compiler>()->compile(format,formatspec.c_str())) {
				pc->playback_format_title(NULL,title,format,NULL,playback_control::display_level_all);
				if ((resetlcd) & (foo_comm_lcdrefresh)) out<<char(4)<<title;					// if lcd reset enabled send eot before text
				else out<< title;
			}
		}

		if (foo_comm_consoledebug) console::formatter() << out.str().c_str()<<" length:"<<out.str().length();						// debug console output
		try {
			pfc::string8 portno_str8;
			foo_comm_comport.get(portno_str8);
			std::string portno_str=portno_str8.toString();
			
			comportParity parity;
			pfc::string8 parity_str8;
			foo_comm_comparity.get(parity_str8);
			std::string parity_str=StringToUpper(parity_str8.toString());
			if (parity_str=="E") parity=PARITY_EVEN;
			else if (parity_str=="O") parity=PARITY_ODD;
			else if (parity_str=="N") parity=PARITY_NONE;
			else if (parity_str=="M") parity=PARITY_MARK;
			else if (parity_str=="S") parity=PARITY_SPACE;
			else parity=PARITY_NONE;

			comportStop stop;
			pfc::string8 stop_str8;
			foo_comm_comstop.get(stop_str8);
			std::string stop_str=StringToUpper(stop_str8.toString());
			if (stop_str=="1") stop=STOP_1;
			else if (stop_str=="2") stop=STOP_2;
			else if (stop_str=="1.5") stop=STOP_1_5;
			else stop=STOP_1;

			ComPort *m_pComPort = new ComPort(portno_str.c_str(), foo_comm_combaud, foo_comm_combits, parity, stop, DTR_LOW, RTS_LOW);  
			// reopens port every time, yeah, its lame but its the easiest way to make code resistent to suspend/hibernate/port disconnects
			// doesn't consume much resources anyway
			
			m_pComPort->txString(out.str().c_str());
			delete m_pComPort;
		}
		catch (...)  // swallow all exceptions, always a good programming practice, who gives a f*** about exceptions anyway...
					 // seriously, we don't want music to stop if there is a problem with com port
		{}
	}
};

static play_callback_static_factory_t<foo_comm> fooplayback;

DECLARE_COMPONENT_VERSION(
  "foo_comm",
  "1.0.0",
  "Foobar COM port display component (c) Milos Protic 2014. Distributed under the terms of GNU LGPL v2.1.\n"
  "COM port class code by Yirkha.\n"
  "\n"
  "Ouputs data specified by outputformat to selected COM port on playback events and during playback on refresh interval.\n"
  "Can be used to output data for lcd displays, volume info for external dac volume control or whatever ...\n"
  "\n"
  "Params: \n"
  "  1. Com port: COM1, COM2, COM3 .... (default: COM1) \n"
  "  2. Baudrate (default:9600)\n"
  "  3. Data bits [4,5,6,7,8]	(default: 8) \n"
  "  4. Stop bit [1,2,1.5]	(default: 1)\n"
  "  5. Parity [n,o,e,m,s]	(default: n)\n"
  "  6. Refresh interval(sec) (default: 1)\n"
  "  7. Output format - defines which data is being sent through com port on every event, uses standard foobar title format syntax, plus %volume% extension which is not provided by standard foobar title format\n"
  "     default (designed for 2x40 lcd): $upper($left($pad(%filename%,$ifequal(%volume%,0,67,59)),$ifequal(%volume%,0,67,59))) $upper($left($pad(%codec%,3),3))/$left($pad(%bitrate%,3),3)/$ifequal(%volume%,0,,$left(%volume%,5)dB/)%playback_time%\n"
  "  8. LCD refresh (default: off) - outputs EOT character that signals lcd.reset prior to transmission, doesn't print EOT during playback to avoid flickering \n"
  "  9. Console debug output (default: off)\n"
);